from datetime import datetime, timedelta
from enum import Enum
from dataclasses import dataclass

import numpy as np
import pandas as pd

from toargridding.toarstats_constants import STATISTICS_LIST, ALLOWED_SAMPLING_VALUES
from toargridding.static_metadata import global_cf_attributes, TOARVariable

from typing import Dict

date_created = datetime.utcnow().strftime("%Y-%m-dT%H:%M:%SZ")
#date_created = datetime.now(datetime.UTC).strftime("%Y-%m-dT%H:%M:%SZ") # fix as utcnow will be removed in the future

COORDINATE_VARIABLES = ["latitude", "longitude", "time"]
DATA_VARIABLES = ["mean", "std", "n"]


DataVariables = Enum("DataVariables", DATA_VARIABLES)
Coordinates = Enum("Coordinates", COORDINATE_VARIABLES)
Variables = Enum("Variables", [*DATA_VARIABLES, *COORDINATE_VARIABLES])

TIME_FREQUENCIES = {"daily": "D", "monthly": "M", "yearly": "Y"}  # TODO test


@dataclass
class TimeSample:
    """Sampling in time

    provides conversion into different formats
    
    Attributes:
    start:
        start time point
    end:
        end time point
    sampling:
        temporal aggregation of values, e.g. daily, monthly
    """

    start: datetime
    end: datetime
    sampling: str
    
    @property
    def sampling(self) -> str:
        """sampling for data request

        Sampling, i.e. the period used for the calculation of a parameters within the TOAD DB
        Allows only a limited number of supported sampling durations see toargridding.toarstats_constants.ALLOWED_SAMPLING_VALUES
        """
        return self._sampling
    
    @sampling.setter
    def sampling(self, sampling : str):
        if sampling not in ALLOWED_SAMPLING_VALUES:
            raise ValueError(f"sampling: {sampling} is not in the list of supported samplings for toargridding: {ALLOWED_SAMPLING_VALUES}")
        self._sampling = sampling

    def as_datetime_index(self) -> pd.DatetimeIndex:
        """Conversion to array with all sampled time points
        """
        return pd.period_range(self.start, self.end, freq=self.frequency).to_timestamp()

    @property
    def daterange_option(self) -> str:
        """Conversion of range to a string

        Range is given as [start, end)
        """
        end_with_padding = self.end + timedelta(1)
        return f"{self.start.isoformat()},{end_with_padding.isoformat()}"

    @property
    def frequency(self) -> str:
        """Converts sampling argument from TOAR to Pandas 
        """
        return TIME_FREQUENCIES[self.sampling]

    def as_cf_index(self) -> np.array:
        """conversion to netCDF Climate and Forecast (CF) Metadata Conventions
        
        Calculates the duration in days relative to start time point.
        """
        
        tp = self.as_datetime_index()
        return np.array([ (t - tp[0]).days for t in tp])


@dataclass
class Metadata:
    """Metadata of a request. 
    
    Attributes:
    ----------
    variable:
        support variable of the TOAR data base
    time:
        requested time points
    statistics:
        statistical processing applied by the TOAR database for this request
    """
    variable: TOARVariable
    time: TimeSample
    statistic: str

    @staticmethod
    def construct(standart_name: str, time: TimeSample, stat: str):
        """constructor 
        
        Parameters:
        ----------
        standart_name:
            standard name according to CF
        stat:
            statistical analysis to be done by the TOAR database
        time:
            temporal sampling of this request
        """

        variable = TOARVariable.get(standart_name)
        return Metadata(variable, time, stat)
    
    @property
    def statistic(self) -> str: # TODO make better
        """statistical property for being extracted from the TOAR database

        This can for example be the mean, max, min, median. For a full list see toargridding.toarstat_constants.STATISTICS_LIST

        """
        return self._statistic
    
    @statistic.setter
    def statistic(self, stat : str):
        if stat not in STATISTICS_LIST:
            raise ValueError(f"invalid statistic: {stat}")
        self._statistic = stat

    def get_id(self) -> str:
        """creation of a request specific ID

        For example, used for saving link to results of a request in the cache.
        """
        return f"{self.variable.name}_{self.statistic}_{self.time.daterange_option}_at_{datetime.now().date().isoformat()}"

    def get_title(self) -> str:
        """creation of a title for metadata of a xarray according to the CF convention
        """
        return f"{self.time.sampling} {self.statistic} statistic for {self.variable.standart_name} from {self.time.start} to {self.time.end} aggregated on global grid"

    def get_summary(self) -> str:
        """creation of a descriptive string
         
        At the moment same as title.
        """
        return f"{self.time.sampling} {self.statistic} statistic for {self.variable.standart_name} from {self.time.start} to {self.time.end} aggregated on global grid"


@dataclass
class AnalysisRequestResult:
    """Collection of all results from a gridding analysis

    Parameters
    ----------
    stations_data:
        results of the individual stations
    stations_coords:
        coordinates of the stations
    metadata:
        Metadata of the request.
    """

    stations_data: pd.DataFrame
    stations_coords: pd.DataFrame
    metadata: Metadata


def get_global_attributes(metadata: Metadata) -> Dict:
    """combination of global metadata with request specific values.
    """
    dynamic_cf_attributes = {
        "id": metadata.get_id(),
        "title": metadata.get_title(),
        "summary": metadata.get_summary(),
        "date_created": date_created,
        "history": f"{date_created}: File created by toargridding package using data from toar-analysis service",
        # "geospatial_bounds": 0,  # for polygonal description
        "geospatial_lat_min": -90,  # TODO read from grid description
        "geospatial_lat_max": 90,
        "geospatial_lon_min": -180,
        "geospatial_lon_max": 180,
        "time_coverage_start": metadata.time.start.isoformat(),
        "time_coverage_end": metadata.time.end.isoformat(),
        # "time_coverage_duration": 0, # TODO insert durations
        # "time_coverage_resolution": 0,
    }
    cf_attributes = dynamic_cf_attributes | global_cf_attributes
    return cf_attributes


def get_cf_metadata(variable: Variables, metadata: Metadata | None)-> Dict:
    """get CF meta data for coordinates and requested parameter

    The resulting meta data follow the CF convention: https://cfconventions.org/cf-conventions/cf-conventions.html. 
    
    The resulting dictionary contains the required values for the netCDF files.
    """
    
    match variable.name:
        case Variables.latitude.name:
            cf_metadata = {
                "attributes": {
                    "standard_name": "latitude",
                    "long_name": "latitude in degrees north",
                    "units": "degrees_north",
                    "coverage_type": "coordinate",
                },
                "encoding": {"dtype": "int32"},
            }
        case Variables.longitude.name:
            cf_metadata = {
                "attributes": {
                    "standard_name": "longitude",
                    "long_name": "longitude in degrees east",
                    "units": "degrees_east",
                    "coverage_type": "coordinate",
                },
                "encoding": {"dtype": "int32"},
            }
        case Variables.time.name:
            cf_metadata = {
                "attributes": {
                    "units": f"days since {metadata.time.start.isoformat()}",  # TODO formatting !!
                    "long_name": "time",
                    "standard_name": "time",
                    "calendar": "standard",
                    "coverage_content_type": "coordinate",
                },
                "encoding": {"dtype": "int32"},
            }
        case Variables.mean.name:
            cf_metadata = {
                "attributes": {
                    "units": metadata.variable.units,
                    "long_name": metadata.variable.long_name,
                    "standard_name": metadata.variable.standart_name,
                    "cell_methods": "time: point, latitude: mean, longitude: mean",
                    "coverage_content_type": "physicalMeasurement",
                },
                "encoding": {"dtype": "float32", "_FillValue": np.nan, "zlib": False},
            }
        case Variables.std.name:
            cf_metadata = {
                "attributes": {
                    "units": metadata.variable.units,
                    "long_name": metadata.variable.long_name,
                    # TODO standart name extension vs cell_methods
                    "standard_name": f"{metadata.variable.standart_name} standard_error",
                    "cell_methods": "time: point, latitude: standard_deviation, longitude: standard_deviation",
                    "coverage_content_type": "physicalMeasurement",
                },
                "encoding": {
                    "dtype": "float32",
                    "_FillValue": np.nan,
                    "zlib": False,
                },
            }
        case Variables.n.name:
            cf_metadata = {
                "attributes": {
                    "units": "1",
                    "long_name": f"number of {metadata.variable.long_name} observations in cell",
                    "standard_name": "number_of_observations",
                    "cell_methods": "time: point, latitude: sum, longitude: sum",
                    "coverage_content_type": "auxiliaryInformation",
                },
                "encoding": {
                    "dtype": "int32",
                    "_FillValue": 0,
                    "zlib": False,
                },
            }

    return cf_metadata
